# Copyright 2024 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the PolyBinary """

import os
import pytest

from quark import PolyBinary
from quark.io import hdf5, txt


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE_H5 = "poly_binary.h5"
FILE_TXT = "poly_binary.txt"
FILE_BP = "poly_binary.bp"
FILE_BP_LARGE = "ising2.5-100_5555.bq"
FILENAME_H5_TMP = TEST_DIR + FILE_H5
FILENAME_TXT_TMP = TEST_DIR + FILE_TXT
FILENAME_BP_TMP = TEST_DIR + FILE_BP
FILENAME_H5_FIX = TEST_DIR + "testdata/" + FILE_H5
FILENAME_TXT_FIX = TEST_DIR + "testdata/" + FILE_TXT
FILENAME_BP_FIX = TEST_DIR + "testdata/" + FILE_BP
FILENAME_BP_LARGE_FIX = TEST_DIR + "testdata/" + FILE_BP_LARGE

POLY_FIX = PolyBinary({(("x", 1),): 2.5, (("x", 2), ("y", 0, 0)): 2.0, (): 5.5})
POLYS = [PolyBinary({(1,): 5, (2,): 6, (3, 1): 3, (2, 4): 3, (4,): 1}),
         PolyBinary({(2,): 1, (2, 4): 1, (3, 4): 3, (5,): 7}),
         PolyBinary({(1,): 7, (2,): 2, (3,): 3, (4,): 1}),
         PolyBinary({(1,): 66, (1, 2): 77}),
         PolyBinary({(1,): 1.5, (2,): 2.5})]

POLYS_NOT = [PolyBinary({((1,),): 1.0, ((1, 2),): 2.0, ((1, 2), (1, 2), (1, 3)): 2.0, (): 5.5}),
             PolyBinary({}),
             PolyBinary({(): 55}),
             PolyBinary({(): 55, (1,): 66}),
             PolyBinary({(): 55, (1,): 66, (1, 2): 77})]

def test_load():
    """ test io """
    loaded = hdf5.load(PolyBinary, FILENAME_H5_FIX)
    assert loaded == POLY_FIX

    loaded = txt.load(PolyBinary, FILENAME_TXT_FIX)
    assert loaded == POLY_FIX

    loaded = PolyBinary.load_bp(FILENAME_BP_FIX)
    assert loaded == POLYS[0]

@pytest.mark.parametrize("poly", POLYS + POLYS_NOT)
def test_io(poly):
    """ test io round trip """
    hdf5.save(poly, FILENAME_H5_TMP)
    loaded = hdf5.load(PolyBinary, FILENAME_H5_TMP)
    assert loaded == poly
    os.remove(FILENAME_H5_TMP)

    txt.save(poly, FILENAME_TXT_TMP)
    loaded = txt.load(PolyBinary, FILENAME_TXT_TMP)
    assert loaded == poly
    os.remove(FILENAME_TXT_TMP)

@pytest.mark.parametrize("poly", POLYS)
def test_io_bp(poly):
    """ test io round trip """
    assert not PolyBinary.exists_bp(FILENAME_BP_TMP)
    poly.save_bp(FILENAME_BP_TMP)
    assert PolyBinary.exists_bp(FILENAME_BP_TMP)

    loaded = PolyBinary.load_bp(FILENAME_BP_TMP)
    assert loaded == poly
    os.remove(FILENAME_BP_TMP)

def test_io_bp_not():
    """ test io failure """
    for poly in POLYS_NOT[:3]:
        with pytest.raises(ValueError, match="Can only write"):
            poly.save_bp(FILENAME_BP_TMP)
    for poly in POLYS_NOT[3:]:
        with pytest.warns(UserWarning, match="Offset will be ignored"):
            poly.save_bp(FILENAME_BP_TMP)
    os.remove(FILENAME_BP_TMP)

def test_io_bp_large():
    """ test large instance """
    loaded = PolyBinary.load_bp(FILENAME_BP_LARGE_FIX)
    assert loaded.get_variable_num() == 100
    assert len(loaded) == 4981
    assert loaded.get_coefficient(47, 55) == -1086

def test_bp_exist():
    """ test existence """
    with open("empty_file.txt", mode="w"):
        pass
    assert not PolyBinary.exists_bp("empty_file.txt")
    os.remove("empty_file.txt")

    with open("wrong_format.txt", mode="w", encoding="utf-8") as wrong_format_file:
        # get the corresponding function for writing and call it
        wrong_format_file.write(f"1 1")  # wrong format -> no variables
    assert not PolyBinary.exists_bp("wrong_format.txt")
    os.remove("wrong_format.txt")

    with open("wrong_format.txt", mode="w", encoding="utf-8") as wrong_format_file:
        # get the corresponding function for writing and call it
        wrong_format_file.write(f"1 1 1")  # wrong format -> header is not set
    assert not PolyBinary.exists_bp("wrong_format.txt")
    os.remove("wrong_format.txt")



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the PolyBinary """

import os
import ast
from warnings import warn

from quark import PolyBinary
from . import add


@add.self_method(PolyBinary)
def save_bp(polynomial, filename):
    """
    save polynomial in bp file (only one polynomial per file)

    :param polynomial: the polynomial to be saved
    :param (str) filename: the name of bp file
    """
    if not polynomial.is_quadratic() or not polynomial.is_flat() or polynomial.is_constant():
        raise ValueError("Can only write non-constant, flat and quadratic binary polynomials in this format")
    if polynomial.offset:
        warn("Offset will be ignored")

    with open(filename, mode="w", encoding="utf-8") as bp_file:
        # get the corresponding function for writing and call it
        bp_file.write(f"{max(polynomial.variables)} {len(polynomial.linear) + len(polynomial.quadratic)}\n")
        for (var,), coeff in polynomial.linear.items():
            bp_file.write(f"{var} {var} {coeff}\n")
        for (var1, var2), coeff in polynomial.quadratic.items():
            bp_file.write(f"{var1} {var2} {coeff}\n")

@add.class_method(PolyBinary)
def load_bp(cls, filename):
    """
    load polynomial of the given class from bp file

    :param cls: the type of the polynomial to be loaded
    :param (str) filename: the name of bp file
    """
    with open(filename, "r", encoding="utf-8") as bp_file:
        lines = bp_file.readlines()

    poly_dict = _get_poly_dict(lines)
    loaded_polynomial = cls(poly_dict)
    return loaded_polynomial

@add.class_method(PolyBinary)
def exists_bp(cls, filename):
    """
    check if a binary polynomial exists in the bp file

    :param cls: the type of object to be checked
    :param (str) filename: the name of bp file
    :return: True if there is already a binary polynomial in the file
    """
    if os.path.exists(filename):
        with open(filename, "r", encoding="utf-8") as pb_file:
            lines = pb_file.readlines()
        try:
            poly_dict = _get_poly_dict(lines)
            _ = cls(poly_dict)
            return True
        except ValueError:
            pass
    return False


def _get_poly_dict(lines):
    poly_dict = {}
    max_variable, num_terms = None, None
    for line in lines:
        if line == "\n" or line.startswith("#"):
            continue
        values = line.split(" ")
        if len(values) == 2:
            max_variable, num_terms = (ast.literal_eval(value) for value in values)
        else:
            var1, var2, coeff = (ast.literal_eval(value) for value in values)
            poly_dict[(var1, var2)] = coeff
    variables = set(var for var_tuple in poly_dict for var in var_tuple)
    if not max(variables) == max_variable or not len(poly_dict) == num_terms:
        raise ValueError("Something went wrong when reading the file")
    return poly_dict



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of attributes in HDF5 files """

import os
import pytest
import numpy as np
import h5py

from quark.io import hdf5_attributes


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "attributes.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE
GROUP_NAME = "testgroup_attributes"

ATTRS = {"a1_int":   1,
         "a2_str":   "two",
         "a3_bool":  True,
         "a4_list":  [1, 2],
         "a5_tuple": (1, 2),
         "a6_set":   {1, 2},
         "a7_None":  None}

class ExampleObject:
    """ some test object with data """

    def __init__(self):
        self.a1_int = 1
        self.a2_str = "two"
        self.a3_bool = True
        self.a4_list = [1, 2]
        self.a5_tuple = (1, 2)
        self.a6_set = {1, 2}


def test_write_attributes_dict():
    """ test write_attribute function with dictionary of names and values """
    with h5py.File(FILENAME_TMP, "w") as h5file:
        group = h5file.create_group(GROUP_NAME)
        hdf5_attributes.write_attributes(group, a0_float=1.4, **ATTRS)

        with pytest.raises(ValueError, match="Given value does not have an easy data structure, create dataset instead"):
            hdf5_attributes.write_attributes(group, a10_invalid=[(1, 3), (3, 4)])

        with pytest.raises(ValueError, match="Given value does not have an easy data structure, create dataset instead"):
            hdf5_attributes.write_attributes(group, a10_invalid={1 : 3})

    _check_and_remove_file()

def test_write_attributes_obj():
    """ test write_attribute function with provided object """
    example_object = ExampleObject()

    with h5py.File(FILENAME_TMP, "w") as h5file:
        group = h5file.create_group(GROUP_NAME)
        # for testing overwrite
        hdf5_attributes.write_attribute(group, "a1_int", value=2)
        # tests taking the attributes from the object and additionally the keyword plus its value
        hdf5_attributes.write_attributes(group, example_object, *ATTRS.keys(), a0_float=1.4)
        # test to store an attribute that the object does not have (results in storing None)
        hdf5_attributes.write_attribute(group, "a7_None", example_object)

    _check_and_remove_file()

def test_read_attributes():
    """ test write_attribute function """
    with h5py.File(FILENAME_FIX, "r") as h5file:
        group = h5file[GROUP_NAME]

        loaded_attr = hdf5_attributes.read_attributes(group, *ATTRS.keys())
        assert loaded_attr["a1_int"] == 1
        assert loaded_attr["a2_str"] == "two"
        assert loaded_attr["a3_bool"]
        assert list(loaded_attr["a4_list"]) == [1, 2]
        assert tuple(loaded_attr["a5_tuple"]) == (1, 2)
        assert set(loaded_attr["a6_set"]) == {1, 2}
        assert loaded_attr["a7_None"] is None

        not_existing_no_check = hdf5_attributes.read_attributes(group, "a10_whatever", check_existence=False)
        assert not_existing_no_check["a10_whatever"] is None

        with pytest.raises(ValueError, match="Did not find attribute 'a10_whatever' at group '/" + GROUP_NAME):
            hdf5_attributes.read_attributes(group, "a10_whatever")


def _check_and_remove_file():
    with h5py.File(FILENAME_TMP, "r") as h5file:
        group = h5file[GROUP_NAME]

        for name in ATTRS.keys():
            assert name in group.attrs
        assert "a0_float" in group.attrs

        assert group.attrs["a1_int"] == 1
        assert group.attrs["a2_str"].decode() == "two"
        assert group.attrs["a3_bool"]
        assert list(group.attrs["a4_list"]) == [1, 2]
        assert tuple(group.attrs["a5_tuple"]) == (1, 2)
        assert set(group.attrs["a6_set"]) == {1, 2}
        assert np.isnan(group.attrs["a7_None"])
        assert group.attrs["a0_float"] == 1.4

    os.remove(FILENAME_TMP)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of datasets in HDF5 files """

import os
import pytest
import numpy as np
import h5py

from quark.io import hdf5_datasets


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "datasets.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE
GROUP_NAME = "testgroup_datasets"

SUPPORTED_DATASETS = {
    "empty_list": [],
    "1d_int": [1, 2, 3, 4, 5],
    "1d_float": [1.0, 2.5, 3.333333],
    "1d_mixed": [1, 2.5, 3],
    "1d_strings": ["one", "two", "three"],
    "2d_fixed_len_int": [[1, 2, 3], [4, 5, 6]],
    "2d_fixed_len_float": [[1.5, 2.0, 3.0], [4.0, 5.0, 6.0]],
    "2d_var_len_int": [[1, 2], [1], [], [1, 2, 3]],
    "2d_var_len_int_tuples": [(1, 2), (1,), (), (1, 2, 3)],
    "2d_var_len_numeric": [[1.0, 2.1], [1.8], [], [0, 2, 3.99]],
    "2d_empty_tuple": [()],
    "dict_int_to_int": {1: 2, 2: 2, 3: 3},
    "dict_string_to_float": {"qubits": 16, "annealing_time": 0.001},
    "dict_int_tuple_to_scalar": {(0,): 2.5, (1, 2, 3): 2.0, (): 5.5},
    "dict_int_tuple_to_scalar2": {(): 55, (2,): 66},
    "dict_empty_tuple": {(): 2.5},
    "dict_int_tuple_one_entry": {(1,): 2.5},
    "dict_int_to_string": {1: "x", 2: "y", 50: "test"},
    "empty_dict": {},
    "numpy_array": np.array([1, 2, 3, 4, 5]),
}

NOT_SUPPORTED_DATASETS = {
    "2d_var_len_string": [("x", "0"), ("reduction", "x", "0", "y", "0")],
    "2d_var_len_mixed_type": [("x", 1), ("y", 0, 0)],
    "dict_tuple_key_ints": {((10, 1),): 2.5, ((10, 2), (11, 0, 0)): 2.0, (): 5.5},
    "dict_tuple_key_mixed": {(("x", 1),): 2.5, (("x", 2), ("y", 0, 0)): 2.0, (): 5.5}
}

OTHER_DATASET = [-1.0, -10]

class ExampleObject:
    """ some test object """
    pass


@pytest.mark.parametrize("name, dataset", SUPPORTED_DATASETS.items())
def test_read_write_dataset(name, dataset):
    """
    test the read and write functions in hdf5_datasets

    :param name: the name of the dataset
    :param dataset: the dataset to read and write
    """
    with h5py.File(FILENAME_TMP, "w") as h5file:
        group = h5file.create_group(GROUP_NAME)
        hdf5_datasets.write_dataset(group, name, dataset=dataset)

    with h5py.File(FILENAME_TMP, "r") as h5file:
        group = h5file[GROUP_NAME]
        dataset_from_file = hdf5_datasets.read_dataset(group, name)

    _check_datasets(dataset, dataset_from_file)
    os.remove(FILENAME_TMP)

@pytest.mark.parametrize("name, dataset", NOT_SUPPORTED_DATASETS.items())
def test_not_supported_dataset(name, dataset):
    """
    test the read and write functions in hdf5_datasets

    :param name: the name of the dataset
    :param dataset: the dataset to read and write
    """
    with pytest.raises(NotImplementedError, match="Storing of dataset"):
        with h5py.File(FILENAME_TMP, "w") as h5file:
            group = h5file.create_group(GROUP_NAME)
            hdf5_datasets.write_dataset(group, name, dataset=dataset)

def test_other_datasets():
    """ test write function in hdf5_datasets throwing an error for attributes """
    with pytest.raises(ValueError, match="Given dataset is just a scalar, create attribute instead"):
        with h5py.File(FILENAME_TMP, "w") as h5file:
            hdf5_datasets.write_dataset(h5file, "test_string", dataset="string")

    with h5py.File(FILENAME_TMP, "w") as h5file:
        group = h5file.create_group(GROUP_NAME)
        hdf5_datasets.write_dataset(group, "None", dataset=None)

        with pytest.raises(ValueError, match="Storing of dataset with type"):
            hdf5_datasets.write_dataset(group, "numpy_array_object", dataset=np.array([1, 2, 3, 4, 5], dtype=object))
        with pytest.raises(NotImplementedError, match="Storing of dataset with type list over '<class 'type'>'"):
            hdf5_datasets.write_dataset(group, "numpy_string", dataset=[object, object])
        with pytest.raises(TypeError, match="Dictionary keys are not of homogeneous type"):
            hdf5_datasets.write_dataset(group, "dict_keys_var_len_mixed_type", dataset={("x", 1): 1, "y0": 2})
        with pytest.raises(TypeError, match="Key items are not of homogeneous type"):
            hdf5_datasets.write_dataset(group, "dict_keys_var_len_mixed_type", dataset={("x", 1): 1, ("y", 0, 0): 2})

    with h5py.File(FILENAME_TMP, "r") as h5file:
        group = h5file[GROUP_NAME]
        dataset_from_file = hdf5_datasets.read_dataset(group, "None")
        assert dataset_from_file is None

    os.remove(FILENAME_TMP)

def test_write_read_datasets_obj():
    example_object = ExampleObject()
    for name, ds in SUPPORTED_DATASETS.items():
        example_object.__setattr__(name, ds)


    with h5py.File(FILENAME_TMP, "w") as h5file:
        group = h5file.create_group(GROUP_NAME)
        # test with overwrite
        hdf5_datasets.write_dataset(group, "other", dataset=[])
        # tests taking the datasets from the object and additionally the keyword plus its value
        hdf5_datasets.write_datasets(group, example_object, *SUPPORTED_DATASETS.keys(), other=OTHER_DATASET)

    with h5py.File(FILENAME_TMP, "r") as h5file:
        group = h5file[GROUP_NAME]
        read_datasets = hdf5_datasets.read_datasets(group, "other", *SUPPORTED_DATASETS.keys())

        assert len(read_datasets) == len(SUPPORTED_DATASETS) + 1
        for name in SUPPORTED_DATASETS.keys():
            # checking if they read in datasets are equal to the stored ones
            _check_datasets(getattr(example_object, name), read_datasets[name])
        _check_datasets(OTHER_DATASET, read_datasets["other"])

    os.remove(FILENAME_TMP)

def test_read_datasets():
    with h5py.File(FILENAME_FIX, "r") as h5file:
        group = h5file[GROUP_NAME]
        read_datasets = hdf5_datasets.read_datasets(group, "other", *SUPPORTED_DATASETS.keys())

        assert len(read_datasets) == len(SUPPORTED_DATASETS) + 1
        for name, ds in SUPPORTED_DATASETS.items():
            # checking whether the read in datasets are as we expect them
            _check_datasets(ds, read_datasets[name])
        _check_datasets(OTHER_DATASET, read_datasets["other"])

        with pytest.raises(ValueError, match="Mismatched dictionary dataset"):
            hdf5_datasets.read_dataset(group, "dict_unreadable")


def _check_datasets(dataset, dataset_from_file):
    if isinstance(dataset, dict):
        assert dataset_from_file == dataset
    else:
        assert len(dataset) == len(dataset_from_file)
        assert all(np.array_equal(a, b) for a, b in zip(dataset, dataset_from_file))



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" test ising polynomial module """

import pytest
import numpy as np

from quark import PolyIsing, PolyBinary, Polynomial


X, Y = "x", "y"

P1_DICT = {(1,): 5, (2,): 6, (1, 3): -3, (2, 4): 3, (4,): -1, (): 3}
P2_DICT = {((X, 1), (Y, 2, 3)): -2, ((Y, 1, 2),): 3}
P3_DICT = {(0,): 5, (1,): 6, (2, 0): -3, (1, 3): 0.5, (3,): -1, (0, 1): 0.9, (): 16, (2,): 0}
P1_INV_DICT = {(1,): -5, (2,): -6, (1, 3): -3, (2, 4): 3, (4,): 1, (): 3}
P1_BINARY_DICT = {(): -7, (1,): 16, (2,): 6, (3,): 6, (4,): -8, (1, 3): -12, (2, 4): 12}

POLY1 = PolyIsing(P1_DICT)
POLY2 = PolyIsing(P2_DICT, inverted=True)
POLY3 = PolyIsing(P3_DICT)
POLY1_INV = PolyIsing(P1_INV_DICT, inverted=True)
POLY1_NONE = Polynomial(P1_DICT)
POLY1_BINARY = PolyBinary(P1_BINARY_DICT)

POLY3_LINEAR = [ 5.,  6.,  0., -1.]
POLY3_QUADRATIC = [[ 0. ,  0.9, -3. ,  0. ],
                   [ 0. ,  0. ,  0. ,  0.5],
                   [ 0. ,  0. ,  0. ,  0. ],
                   [ 0. ,  0. ,  0. ,  0. ]]
POLY3_QUAD_DIAG = [[ 5. ,  0.9, -3. ,  0. ],
                   [ 0. ,  6. ,  0. ,  0.5],
                   [ 0. ,  0. ,  0. ,  0. ],
                   [ 0. ,  0. ,  0. , -1. ]]

FROM_RIGHT_WARNING = "PolyIsing will be cast to Polynomial when added or multiplied from the right"


def check_polys_ising(received, expected):
    """ check inverted Ising polynomials """
    # pylint: disable=unidiomatic-typecheck
    assert type(received) is PolyIsing
    assert received == expected
    assert received.is_inverted() == expected.is_inverted()


def test_init():
    """ test initialization """
    poly = PolyIsing({(1, 1, 1): 5, (2, 4, 4, 4, 4): 6, (3, 1): -3, (2, 4): 3, (4,): -1, (2, 2): 3})
    assert POLY1 == poly
    assert P1_DICT == dict(poly)

    poly = PolyIsing({((Y, 2, 3), (X, 1), (Y, 2, 3), (Y, 2, 3)): -2, ((Y, 1, 2),): 3}, inverted=True)
    assert POLY2 == poly
    assert P2_DICT == dict(poly)

def test_repr():
    """ test string representation """
    assert repr(POLY1) == "PolyIsing({(): 3, (1,): 5, (2,): 6, (4,): -1, (1, 3): -3, (2, 4): 3})"
    assert repr(POLY1_INV) == "PolyIsing({(): 3, (1,): -5, (2,): -6, (4,): 1, (1, 3): -3, (2, 4): 3})"

def test_str():
    """ test string representation """
    assert str(POLY1) == '+3 +5 s1 +6 s2 -1 s4 -3 s1 s3 +3 s2 s4'

def test_not_eq():
    """ test unequal """
    assert POLY1 != "nonsense"
    assert POLY1 != PolyBinary(P1_DICT)
    assert PolyBinary(P1_DICT) != POLY1

def test_invert():
    """ test inversion """
    # pylint: disable=protected-access
    exp_poly = PolyIsing({(1,): -5, (2,): -6, (1, 3): -3, (2, 4): 3, (4,): 1, (): 3}, inverted=True)
    inverted_poly = POLY1.invert()
    assert inverted_poly == exp_poly
    assert inverted_poly._inverted
    assert inverted_poly.is_inverted()

    exp_poly = PolyIsing({((X, 1), (Y, 2, 3)): -2, ((Y, 1, 2),): -3})
    inverted_poly = PolyIsing(P2_DICT, inverted=True).invert()
    assert inverted_poly == exp_poly
    assert not inverted_poly._inverted
    assert not inverted_poly.is_inverted()

def test_add():
    """ test addition """
    # pylint: disable=unidiomatic-typecheck
    exp_poly = PolyIsing({(1,): 5, (2,): 6, (1, 3): -3, (2, 4): 3, (4,): -1, (): 4})
    sum_poly = PolyIsing(P1_DICT)
    sum_poly += 1
    check_polys_ising(sum_poly, exp_poly)
    check_polys_ising(POLY1 + 1, exp_poly)
    check_polys_ising(1 + POLY1, exp_poly)

    exp_poly = PolyIsing({(1,): 5, (2,): 6, (1, 3): -3, (2, 4): 3, (): 3})
    sum_poly = PolyIsing(P1_DICT)
    sum_poly += PolyIsing({(4,): 1})
    check_polys_ising(sum_poly, exp_poly)
    check_polys_ising(POLY1 + PolyIsing({(4,): 1}), exp_poly)
    check_polys_ising(PolyIsing({(4,): 1}) + POLY1, exp_poly)

    sum_poly = PolyIsing(P1_DICT)
    sum_poly += Polynomial({(4,): 1})
    check_polys_ising(sum_poly, exp_poly)
    check_polys_ising(POLY1 + Polynomial({(4,): 1}), exp_poly)

    with pytest.warns(UserWarning, match=FROM_RIGHT_WARNING):
        sum_poly = Polynomial({(4,): 1}) + POLY1
        assert sum_poly == exp_poly
        assert type(sum_poly) is Polynomial

    with pytest.warns(UserWarning, match=FROM_RIGHT_WARNING):
        sum_poly = Polynomial({(4,): 1})
        sum_poly += POLY1
        assert sum_poly == exp_poly
        assert type(sum_poly) is Polynomial

    with pytest.raises(ValueError, match="Cannot add two PolyIsings with different inversions"):
        _ = PolyIsing(P1_DICT, inverted=True) + POLY1
    with pytest.raises(ValueError, match="Cannot add two PolyIsings with different inversions"):
        _ = POLY1 + PolyIsing(P1_DICT, inverted=True)
    with pytest.raises(ValueError, match="Cannot add PolyIsing and PolyBinary"):
        _ = POLY1 + PolyBinary(P1_DICT)

def test_mul():
    """ test multiplication """
    # pylint: disable=unidiomatic-typecheck
    exp_poly = PolyIsing({(1, 4): 5, (2, 4): 6, (1, 3, 4): -3, (2,): 3, (): -1, (4,): 3})
    prod_poly = PolyIsing(P1_DICT)
    prod_poly *= PolyIsing({(4,): 1})
    check_polys_ising(prod_poly, exp_poly)
    check_polys_ising(POLY1 * PolyIsing({(4,): 1}), exp_poly)
    check_polys_ising(PolyIsing({(4,): 1}) * POLY1, exp_poly)

    exp_poly = PolyIsing({((X, 1), (Y, 2, 3), (Y, 1, 2)): -12, (): 13}, inverted=True)
    prod_poly = POLY2 * POLY2
    check_polys_ising(prod_poly, exp_poly)

    factor_poly = PolyIsing({((X, 1),): 2}, inverted=True)
    exp_poly = PolyIsing({((Y, 2, 3),): -4, ((X, 1), (Y, 1, 2)): 6}, inverted=True)
    prod_poly = PolyIsing(P2_DICT, inverted=True)
    prod_poly *= factor_poly
    check_polys_ising(prod_poly, exp_poly)
    check_polys_ising(POLY2 * factor_poly, exp_poly)
    check_polys_ising(factor_poly * POLY2, exp_poly)

    prod_poly = PolyIsing(P2_DICT, inverted=True)
    prod_poly *= factor_poly
    check_polys_ising(prod_poly, exp_poly)
    check_polys_ising(POLY2 * factor_poly, exp_poly)

    exp_poly = Polynomial({((Y, 2, 3), (X, 1), (X, 1)): -4, ((X, 1), (Y, 1, 2)): 6})
    with pytest.warns(UserWarning, match=FROM_RIGHT_WARNING):
        prod_poly = Polynomial({((X, 1),): 2}) * POLY2
        assert prod_poly == exp_poly
        assert type(prod_poly) is Polynomial

    with pytest.warns(UserWarning, match=FROM_RIGHT_WARNING):
        prod_poly = Polynomial({((X, 1),): 2})
        prod_poly *= POLY2
        assert prod_poly == exp_poly
        assert type(prod_poly) is Polynomial

    with pytest.raises(ValueError, match="Cannot multiply two PolyIsings with different inversions"):
        _ = PolyIsing(P2_DICT, inverted=False) * POLY2

    with pytest.raises(ValueError, match="Cannot multiply two PolyIsings with different inversions"):
        _ = POLY2 * PolyIsing(P2_DICT, inverted=False)

    with pytest.raises(ValueError, match="Cannot multiply PolyIsing and PolyBinary"):
        _ = POLY2 * PolyBinary(P2_DICT)

def test_copy():
    """ test copying """
    # pylint: disable=unidiomatic-typecheck
    poly1_copy = POLY1.copy()
    assert poly1_copy == POLY1
    assert dict(poly1_copy) == dict(POLY1)
    assert poly1_copy.is_inverted() == POLY1.is_inverted()
    assert type(poly1_copy) is PolyIsing

    poly2_copy = POLY2.copy()
    assert poly2_copy == POLY2
    assert dict(poly2_copy) == dict(POLY2)
    assert poly2_copy.is_inverted() == POLY2.is_inverted()
    assert type(poly1_copy) is PolyIsing

def test_remove_zero_coefficients():
    """ test removal of monomials with zero coefficients """
    # pylint: disable=unidiomatic-typecheck
    poly_with_zero = {(3,): 0, (2, 3): 0}
    poly_with_zero.update(P1_DICT)
    poly = PolyIsing(poly_with_zero)
    assert dict(poly) != dict(POLY1)
    assert poly == POLY1

    poly_nonzero = poly.remove_zero_coefficients()
    assert poly_nonzero == POLY1
    assert dict(poly_nonzero) == dict(POLY1)
    assert type(poly_nonzero) is PolyIsing

def test_preprocess_minimize():
    """ test preprocessing """
    exp_poly = PolyIsing({(): -15})
    exp_variables = {1 : -1, 2 : -1, 3 : -1, 4 : 1}
    preprocessed_poly, preprocessed_variables = POLY1.preprocess_minimize()
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables

    exp_poly = PolyIsing({((X, 1), (Y, 2, 3)): -2, (): -3}, inverted=True)
    exp_variables = {(Y, 1, 2) : -1}
    preprocessed_poly, preprocessed_variables = POLY2.preprocess_minimize()
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables

    poly = PolyIsing({(1,) : -1, (0, 1) : -1, (0,) : 2, (0, 2) : -1, (2,) : -1})
    exp_variables = {0: -1, 1: 1, 2: 1}
    preprocessed_poly, preprocessed_variables = poly.preprocess_minimize()
    assert preprocessed_poly == -2
    assert preprocessed_variables == exp_variables

    poly = PolyIsing({(0,): 10, (1,): 2, (2,): -8, (3,): 5, (4,): -7, (5,): 4, (0, 1): 3, (0, 2): -4, (0, 3): -3,
                      (1, 3): -8, (1, 4): -7, (2, 3): 8, (2, 5): 1, (3, 4): 3, (3, 5): 2})
    exp_variables = {0: -1, 5: -1}
    exp_poly = PolyIsing({(): -14, (1,): -1, (2,): -5, (3,): 6, (4,): -7, (1, 3): -8, (1, 4): -7, (2, 3): 8, (3, 4): 3})
    preprocessed_poly, preprocessed_variables = poly.preprocess_minimize()
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables

    exp_variables = {5: -1}
    exp_poly = PolyIsing({(): -4, (0,): 10, (1,): 2, (2,): -9, (3,): 3, (4,): -7, (0, 1): 3, (0, 2): -4, (0, 3): -3,
                          (1, 3): -8, (1, 4): -7, (2, 3): 8, (3, 4): 3})
    preprocessed_poly, preprocessed_variables = poly.preprocess_minimize(unambitious=True)
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables

def test_to_binary():
    """ test conversion to binary polynomial """
    poly_ising = PolyIsing.read_from_string("32 + 13 x1 + 9 x2 + 3 x3 - 5 x4 + 3 x1 x3 - 3 x2 x4")
    poly_ising_inv = poly_ising.invert()
    exp_binary_poly = 4 * PolyBinary.read_from_string("5 x1 + 6 x2 + 3 x1 x3 - x4 - 3 x2 x4 + 3")
    assert poly_ising.to_binary() == exp_binary_poly
    assert poly_ising_inv.to_binary() == exp_binary_poly

    poly_ising_inv = POLY1.invert()
    exp_binary_poly = PolyBinary({(): -7, (1,): 16, (1, 3): -12, (2,): 6, (2, 4): 12, (3,): 6, (4,): -8})
    assert POLY1.to_binary() == exp_binary_poly
    assert poly_ising_inv.to_binary() == exp_binary_poly

    poly_ising_inv = POLY2.invert()
    exp_binary_poly = PolyBinary({(): 1, ((X, 1),): 4, ((X, 1), (Y, 2, 3)): -8, ((Y, 1, 2),): -6,
                                  ((Y, 2, 3),): 4})
    assert POLY2.to_binary() == exp_binary_poly
    assert poly_ising_inv.to_binary() == exp_binary_poly

def test_from_unknown_poly_to_ising_no_inv():
    """ test conversion to ising polynomial without inverting ising and without setting ising to certain inversion """
    assert PolyIsing.from_unknown_poly(POLY1) == POLY1
    assert PolyIsing.from_unknown_poly(POLY1_INV) == POLY1_INV

    poly = PolyIsing.from_unknown_poly(POLY1_NONE)
    assert poly == POLY1_NONE
    assert isinstance(poly, PolyIsing)
    assert not poly.is_inverted()

    assert PolyIsing.from_unknown_poly(POLY1_BINARY) == POLY1

def test_unknown_poly_to_ising_invert():
    """ test conversion to ising polynomial with inverting ising """
    assert PolyIsing.from_unknown_poly(POLY1, invert=True) == POLY1_INV
    assert PolyIsing.from_unknown_poly(POLY1_INV, invert=True) == POLY1

    poly = PolyIsing.from_unknown_poly(POLY1_NONE, invert=True)
    assert poly == POLY1_NONE
    assert isinstance(poly, PolyIsing)
    assert poly.is_inverted()

    assert PolyIsing.from_unknown_poly(POLY1_BINARY, invert=True) == POLY1_INV

def test_unknown_poly_to_ising_set_inverted():
    """ test conversion to ising polynomial with setting ising to certain inversion """
    assert PolyIsing.from_unknown_poly(POLY1, inverted=False) == POLY1
    assert PolyIsing.from_unknown_poly(POLY1_INV, inverted=False) == POLY1
    assert PolyIsing.from_unknown_poly(POLY1, inverted=True) == POLY1_INV
    assert PolyIsing.from_unknown_poly(POLY1_INV, inverted=True) == POLY1_INV

    poly = PolyIsing.from_unknown_poly(POLY1_NONE, inverted=True)
    assert poly == POLY1_NONE
    assert isinstance(poly, PolyIsing)
    assert poly.is_inverted()

    poly = PolyIsing.from_unknown_poly(POLY1_NONE, inverted=False)
    assert poly == POLY1_NONE
    assert isinstance(poly, PolyIsing)
    assert not poly.is_inverted()

    assert PolyIsing.from_unknown_poly(POLY1_BINARY, inverted=True) == POLY1_INV
    assert PolyIsing.from_unknown_poly(POLY1_BINARY, inverted=False) == POLY1

def test_from_unknown_poly_to_ising_error():
    """ test conversion to ising polynomial error """
    with pytest.raises(ValueError, match="Choose either to set certain inversion"):
        PolyIsing.from_unknown_poly(POLY1, inverted=True, invert=True)

def test_affine_transform():
    """ test affine transformation """
    # pylint: disable=unidiomatic-typecheck
    exp_poly = Polynomial({(): 43, (1,): -14, (2,): 36, (3,): -24, (4,): 22, (1, 3): -12, (2, 4): 12})
    transformed_poly = POLY1.affine_transform(2, 4)
    assert transformed_poly == exp_poly
    assert type(transformed_poly) is Polynomial

def test_evaluate():
    """ test evaluation of Ising polynomial """
    # pylint: disable=unidiomatic-typecheck
    var_assignment = {(X, 1): -1, (Y, 2, 3): 1}
    evaluated_poly = POLY2.evaluate(var_assignment)
    exp_poly = PolyIsing({(): 2, ((Y, 1, 2),): 3}, inverted=True)
    check_polys_ising(evaluated_poly, exp_poly)

    var_assignment = {(X, 1) : -1, (Y, 2, 3) : 1, (Y, 1, 2) : -1}
    evaluated_poly = POLY2.evaluate(var_assignment)
    assert evaluated_poly == -1

    with pytest.raises(ValueError, match="Can only assign numbers or polynomials"):
        var_assignment = {(X, 1): PolyIsing({((X, 1),): 2, (): 3})}
        POLY2.evaluate(var_assignment)

    var_assignment = {(X, 1): Polynomial({(("z", 1),): 2, (): 3})}
    exp_poly = Polynomial({(("z", 1), (Y, 2, 3)): -4, ((Y, 1, 2),): 3, ((Y, 2, 3),): -6})
    evaluated_poly = POLY2.evaluate(var_assignment)
    assert type(evaluated_poly) is Polynomial
    assert evaluated_poly == exp_poly

def test_replace_variables():
    """ test replacement of variables """
    replacement = {(X, 1) : "a", (Y, 2, 3) : "b", (Y, 1, 2) : "c"}
    exp_poly = PolyIsing({("a", "b"): -2, ("c",): 3}, inverted=True)
    replaced_poly = POLY2.replace_variables(replacement)
    check_polys_ising(replaced_poly, exp_poly)

def test_get_rounded():
    """ test rounding """
    poly = PolyIsing({((X, 1), (Y, 2, 3)): -2.3, ((Y, 1, 2),): 2.9}, inverted=True)
    rounded_poly = poly.round(0)
    check_polys_ising(rounded_poly, POLY2)

def test_get_matrix_representation():
    """ test standard matrix representation """
    with pytest.warns(UserWarning, match="Constant offset of 16 is dropped"):
        linear, quadratic = POLY3.get_matrix_representation()
    assert np.array_equal(linear, POLY3_LINEAR)
    assert np.array_equal(quadratic, POLY3_QUADRATIC)

def test_get_from_matrix_representation():
    """ test conversion from matrix representation """
    poly = PolyIsing.get_from_matrix_representation(POLY3_LINEAR, POLY3_QUADRATIC)
    assert poly == POLY3 - 16

    poly = PolyIsing.get_from_matrix_representation(POLY3_QUAD_DIAG)
    exp_poly = PolyIsing({(2, 0): -3, (1, 3): 0.5, (0, 1): 0.9, (): 10})
    assert poly == exp_poly



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the ObjectiveTerms """

from quark import ObjectiveTerms
from . import add, hdf5, hdf5_datasets
from .poly_utils import write_poly_with_type, read_poly_with_type


OBJECTIVE_TERMS        = "objective_terms"
CONSTRAINT_TERMS_NAMES = "constraint_terms_names"


add.save_load_exists(ObjectiveTerms)

@add.static_method(ObjectiveTerms)
def get_identifying_attributes():
    """ the attributes that identify an object of the class """
    return ["name"]

@add.self_method(ObjectiveTerms)
def write_hdf5(objective_terms, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (ObjectiveTerms) objective_terms: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    hdf5_datasets.write_dataset(group, CONSTRAINT_TERMS_NAMES, objective_terms)
    for name, poly in objective_terms.items():
        write_poly_with_type(group, poly, name)

@add.static_method(ObjectiveTerms)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    constraint_terms_names = hdf5_datasets.read_dataset(group, CONSTRAINT_TERMS_NAMES, check_existence=False)
    objective_terms = {}
    for name in group:
        if hdf5.is_subgroup(group, name):
            objective_terms[name] = read_poly_with_type(group, name)
    return {OBJECTIVE_TERMS: objective_terms, CONSTRAINT_TERMS_NAMES: constraint_terms_names}



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for Embedding """

from math import prod
from functools import cached_property
from itertools import combinations, product
import networkx as nx

from .hardware_adjacency import HardwareAdjacency


COMPLETE = "complete"
ARE_NEIGHBORED = "are_neighbored"

ERROR_MAPPINGS  = "Either provide mapping from variables to nodes or to edges"
ERROR_DIFFERENT = "The var_edges_map cannot contain different variables than var_nodes_map"
ERROR_CONNECTED = "Cannot check connectivity without mapping from variables to edges " \
                  "either embedding should have var_edges_map or an hwa should be provided"
ERROR_VALID     = "Cannot judge validity without mapping from couplings to edges"
ERROR_NODES     = "Each variable should map to at least one node"


class Embedding:
    """
    An embedding is a mapping from variables and couplings of the original problem to the nodes and edges of a hardware
    graph, defining the embedding subgraphs.
    """

    def __init__(self, var_nodes_map=None, coupling_edges_map=None, var_edges_map=None, name=None, index=None):
        """
        initialize Embedding object,
        need at least either var_nodes_map or var_edges_map

        :param (dict or list[list] or None) var_nodes_map: mapping of variables to nodes in the hardware graph,
                                                           if not given, it is extracted from var_edges_map
        :param (dict or None) coupling_edges_map: mapping of couplings to edges in the hardware graph
        :param (dict or list[list] or None) var_edges_map: mapping of variables to edges in the hardware graph
        :param (str or None) name: identifying name to differ between several embeddings
        :param (int or None) index: identifying index to differ between several embeddings with the same name
        """
        if not var_nodes_map and not var_edges_map:
            raise ValueError(ERROR_MAPPINGS)
        if var_edges_map and not var_nodes_map:
            var_nodes_map = {var : sorted(set(node for edge in edges for node in edge))
                             for var, edges in var_edges_map.items()}

        self.name = name
        self.index = index
        self.var_nodes_map = var_nodes_map if isinstance(var_nodes_map, dict) else dict(enumerate(var_nodes_map))
        self.var_edges_map = var_edges_map if isinstance(var_edges_map, dict) else dict(enumerate(var_edges_map or []))
        self.coupling_edges_map = {tuple(sorted(coupling)): edges
                                   for coupling, edges in (coupling_edges_map or {}).items()}
        self.check_consistency()

    def __eq__(self, other):
        if isinstance(other, Embedding):
            return self.var_nodes_map == other.var_nodes_map \
                   and self.var_edges_map == other.var_edges_map \
                   and self.coupling_edges_map == other.coupling_edges_map \
                   and self.name == other.name \
                   and self.index == other.index
        return False

    def __repr__(self):
        """ get nice string representation """
        return self.__class__.__name__ + f"({repr(self.var_nodes_map)}, {repr(self.name)}, {self.index})"

    def __str__(self):
        """ get nice human-readable string representation of the objective """
        return "\n".join(f"{var} \t-> {nodes}" for var, nodes in self.var_nodes_map.items())

    def check_consistency(self):
        """ check if the data is consistent """
        if self.var_edges_map and set(self.var_edges_map.keys()) != set(self.var_nodes_map.keys()):
            raise ValueError(ERROR_DIFFERENT)
        if any(not nodes for nodes in self.var_nodes_map.values()):
            raise ValueError(ERROR_NODES)

    @cached_property
    def nodes(self):
        """ all nodes used in the embedding """
        return sorted(set(node for nodes in self.var_nodes_map.values() for node in nodes))

    @cached_property
    def edges(self):
        """ all edges used in the embedding """
        edges = set()
        if self.var_edges_map:
            edges.update(edge for edges in self.var_edges_map.values() for edge in edges)
        if self.coupling_edges_map:
            edges.update(edge for edges in self.coupling_edges_map.values() for edge in edges)
        return sorted(edges)

    @cached_property
    def max_map_size(self):
        """ the maximum size (number of nodes) of the embedding subgraph of some variable """
        return max((len(nodes) for nodes in self.var_nodes_map.values()), default=0)

    @classmethod
    def get_from_hwa(cls, var_nodes_map, hwa, name=None, index=None):
        """
        initialize embedding with information calculated from the hardware graph

        :param (dict or list[list]) var_nodes_map: mapping of variables to nodes in the hardware graph
        :param (list[tuples]) hwa: hardware adjacency defining the hardware graph to which the embedding should fit
        :param (str or None) name: identifying name to differ between several embeddings,
                                   by default the name of the HWA
        :param (int or None) index: identifying index to differ between several embeddings with the same name
        :return: the embedding compatible to the hwa
        """
        var_nodes_map = var_nodes_map if isinstance(var_nodes_map, dict) else dict(enumerate(var_nodes_map))
        var_edges_map = get_var_edges_map(var_nodes_map, hwa)
        coupling_edges_map = get_coupling_edges_map(var_nodes_map, hwa)
        name = name or (hwa.name if isinstance(hwa, HardwareAdjacency) else None)
        return cls(var_nodes_map, coupling_edges_map, var_edges_map, name=name, index=index)

    def is_valid(self, couplings=None, hwa=None):
        """
        check if the embedding is valid in general
        and, if given, also checks if there is at least one edge in the hardware graph for each coupling

        :param (set[tuples] or list[tuples] or str or None) couplings: list of couplings of the original problem
        :param (list[tuples] or HardwareAdjacency) hwa: hardware adjacency defining the hardware graph
        :return: True if the embedding is valid
        """
        return self.are_all_node_sets_disjoint() \
               and self.are_all_subgraphs_connected(hwa) \
               and self.are_all_edges_valid() \
               and (couplings is None or self.are_all_couplings_mapped(couplings, hwa))

    def are_all_node_sets_disjoint(self):
        """
        check if the node sets do have nodes in common

        :return: True if all node sets are disjoint
        """
        for set1, set2 in combinations(self.var_nodes_map.values(), 2):
            if not set(set1).isdisjoint(set2):
                return False
        return True

    def are_all_subgraphs_connected(self, hwa=None):
        """
        check if all embeddings form a connected subgraph

        :param (list[tuples] or HardwareAdjacency) hwa: hardware adjacency defining the hardware graph
        :return: True if all embedding subgraphs are connected
        """
        if not self.var_edges_map and not hwa:
            raise ValueError(ERROR_CONNECTED)
        var_edges_map = self.var_edges_map or get_var_edges_map(self.var_nodes_map, hwa)

        for variable, edges in var_edges_map.items():
            assert self.var_nodes_map[variable], "there should not be empty nodes maps"
            # in turn, embeddings consisting of a single node, do not have edges, but are validly 'connected'
            if len(self.var_nodes_map[variable]) > 1:
                subgraph = nx.Graph()
                subgraph.add_edges_from(edges)
                if not subgraph or not nx.algorithms.is_connected(subgraph):
                    return False
        return True

    def are_all_edges_valid(self):
        """
        check if all edges for a coupling indeed connect the nodes sets of the corresponding variables

        :return: True if all edges connect the node sets
        """
        for coupling, edges in self.coupling_edges_map.items():
            for edge in edges:
                if not (edge[0] in self.var_nodes_map[coupling[0]] and edge[1] in self.var_nodes_map[coupling[1]]) and \
                        not (edge[1] in self.var_nodes_map[coupling[0]] and edge[0] in self.var_nodes_map[coupling[1]]):
                    return False
        return True

    def are_all_couplings_mapped(self, couplings, hwa=None):
        """
        check if there is at least one edge in the hardware graph for each coupling

        :param (set[tuples] or list[tuples] or str) couplings: set of couplings of the original problem
        :param (list[tuples] or HardwareAdjacency) hwa: hardware adjacency defining the hardware graph
        :return: True if all couplings do have a counterpart in the hardware
        """
        if not self.coupling_edges_map and not hwa:
            raise ValueError(ERROR_VALID)
        coupling_edges_map = self.coupling_edges_map or get_coupling_edges_map(self.var_nodes_map, hwa)

        if couplings == COMPLETE:
            couplings = combinations(self.var_nodes_map.keys(), 2)

        return all(coupling_edges_map.get(tuple(sorted(coupling)), []) for coupling in couplings)

    def get_vars_with_broken_embedding(self, node_assignments):
        """
        get those variables whose mapped nodes are not synchronized

        :param (dict) node_assignments: mapping of nodes to values
        :return: the list of variables
        """
        var_scores = self.get_var_value_scores(node_assignments)
        return [var for var, score in var_scores if len(score) > 1]

    def de_embed(self, node_assignments, single_var_threshold=None, total_threshold=None, return_score=False):
        """
        get the best assignments of the original variables from the node assignments
        based on majority vote if there are broken node sets and the thresholds are unset,
        otherwise all possible configurations are returned,
        in the order of their score until the threshold is reached,
        both thresholds can be combined, where the strongest in each case will apply

        :param (dict) node_assignments: mapping of nodes to values
        :param (float or None) single_var_threshold: the threshold which the score of a single variable assignment
                                                     should exceed, by default None, i.e. no restriction on the single
                                                     variable scores is used, if set, will return all assignments up to
                                                     this threshold, a value of 1 means no broken embeddings are
                                                     accepted, while 0 means all
        :param (float or None) total_threshold: the threshold which the total score of the variable assignments should
                                                exceed, by default None, i.e. no restriction on the total score is used,
                                                if used, will return all assignments up to this threshold, a value of 1
                                                means no broken embeddings are accepted, while 0 means all
        :param (bool) return_score: if True, also get the score(s) of the corresponding variable assignments
        :return: the best de-embedded variable assignment for the original variables if no thresholds are used,
                 otherwise all variable assignments up to the threshold(s)
        """
        var_scores = self.get_var_value_scores(node_assignments)
        best_score = prod(value_scores[0][1] for _, value_scores in var_scores)
        best_assignment = {var : value_scores[0][0] for var, value_scores in var_scores}
        if single_var_threshold is None and total_threshold is None:
            return (best_assignment, best_score) if return_score else best_assignment

        worst_in_best_score = var_scores[-1][1][0][1]  # last variable with worst score in best assignment
        single_var_threshold = single_var_threshold or 0
        total_threshold = total_threshold or 0
        if worst_in_best_score < single_var_threshold or best_score < total_threshold:
            return ([], best_score) if return_score else []

        assignments_scores = get_assignments_scores(var_scores, single_var_threshold, total_threshold)
        if return_score:
            return assignments_scores
        return [assignment for assignment, _ in assignments_scores]

    def get_var_value_scores(self, node_assignments):
        """
        extract the possible assignments for each original variable individually
        and the corresponding score of this assignment

        :param (dict) node_assignments: mapping of nodes to values
        :return: the dictionary of original variables to dictionaries with assignments to vote
        """
        var_assignments =  {var : [node_assignments[node] for node in nodes if node in node_assignments]
                            for var, nodes in self.var_nodes_map.items()}
        scores = {var : {value : values.count(value) / len(values) for value in set(values)}
                  for var, values in var_assignments.items()}
        scores = {var : sorted(values.items(), key=lambda x: x[1], reverse=True)
                  for var, values in scores.items() if values}
        scores = sorted(scores.items(), key=lambda x: x[1][0][1], reverse=True)
        return scores


def get_var_edges_map(var_nodes_map, hwa):
    """
    get the edges in the embedding subgraph for each variable

    :param (dict or list[list]) var_nodes_map: mapping of variables to nodes in the hardware graph
    :param (list[tuples]) hwa: hardware adjacency defining the hardware graph
    :return: the mapping of variables to edges in the hardware graph
    """
    return {var: get_edges_among(nodes, hwa) for var, nodes in var_nodes_map.items()}

def get_edges_among(nodes, hwa):
    """
    get the edges in the embedding subgraph defined by the given nodes

    :param (set or list) nodes: set of nodes in the hardware graph
    :param (list[tuples] or HardwareAdjacency) hwa: hardware adjacency defining the hardware graph
    :return: the corresponding edges in the hardware graph
    """
    are_neighbored = _get_are_neighbored_func(hwa)
    return sorted(tuple(sorted((n1, n2))) for n1, n2 in combinations(nodes, 2) if are_neighbored(n1, n2))

def get_coupling_edges_map(var_nodes_map, hwa):
    """
    get the edges connecting the embedding subgraphs of each variable pair

    :param (dict or list[list]) var_nodes_map: mapping of variables to nodes in the hardware graph
    :param (list[tuples] or HardwareAdjacency) hwa: hardware adjacency defining the hardware graph
    :return: the mapping of couplings to edges in the hardware graph
    """
    return {tuple(sorted((n1, n2))): get_edges_between(var_nodes_map[n1], var_nodes_map[n2], hwa) for
            n1, n2 in combinations(var_nodes_map.keys(), 2)}

def get_edges_between(nodes1, nodes2, hwa):
    """
    get the edges connecting the embedding subgraphs defined by the two node sets

    :param (set or list) nodes1: first set of nodes in the hardware graph
    :param (set or list) nodes2: second set of nodes in the hardware graph
    :param (list[tuples] or HardwareAdjacency) hwa: hardware adjacency defining the hardware graph
    :return: the edges in the hardware graph connecting the corresponding embedding subgraphs
    """
    are_neighbored = _get_are_neighbored_func(hwa)
    return sorted(tuple(sorted((n1, n2))) for n1, n2 in product(nodes1, nodes2) if are_neighbored(n1, n2))

def _get_are_neighbored_func(hwa):
    if hasattr(hwa, ARE_NEIGHBORED):
        return hwa.are_neighbored
    return lambda node1, node2: (node1, node2) in hwa or (node2, node1) in hwa


def get_assignments_scores(variables_values_scores, single_var_threshold, total_threshold):
    """
    get the valid assignments according to their scores

    :param (list) variables_values_scores: the single scores for assigning the variables to certain values
    :param (float) single_var_threshold: the threshold which the score of a single variable assignment should exceed
    :param (float) total_threshold: the threshold which the total score of the variable assignments should exceed
    :return:
    """
    assignments_scores = [({}, [])]
    for variable, var_value_scores in variables_values_scores:
        for value, var_value_score in var_value_scores:
            if var_value_score < single_var_threshold:
                break
            _add_var_assignment(assignments_scores, variable, value, var_value_score, total_threshold)

    return [(assignment, prod(assignment_scores))
            for assignment, assignment_scores in assignments_scores if len(assignment) == len(variables_values_scores)]

def _add_var_assignment(assignments_scores, variable, value, var_value_score, total_threshold):
    """
    add the variable to the already collected assignments

    :param assignments_scores: the already collected variable assignments
    :param variable: the variable to be added
    :param value: the value which the variable is assigned to
    :param var_value_score: the score of the single variable assignment
    :param total_threshold: the threshold which the total score of the variable assignments should exceed
    """
    new_assignments_scores = []
    for assignment, assignment_single_scores in assignments_scores:
        if not variable in assignment:
            if prod(assignment_single_scores) * var_value_score < total_threshold:
                break
            assignment[variable] = value
            assignment_single_scores.append(var_value_score)
        else:
            new_scores = assignment_single_scores.copy()
            new_scores[-1] = var_value_score
            if prod(new_scores) < total_threshold:
                break
            new_assignment = assignment.copy()
            new_assignment[variable] = value
            new_assignments_scores.append((new_assignment, new_scores))

    assignments_scores += new_assignments_scores



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the ConstraintBinary """

from quark import ConstraintBinary
from . import add, hdf5_attributes
from .poly_utils import write_poly_with_type, read_poly_with_type


POLYNOMIAL  = "polynomial"
LOWER_BOUND = "lower_bound"
UPPER_BOUND = "upper_bound"


add.save_load_exists(ConstraintBinary)

@add.self_method(ConstraintBinary)
def write_hdf5(constraint_binary, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (ConstraintBinary) constraint_binary: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    hdf5_attributes.write_attributes(group, constraint_binary, LOWER_BOUND, UPPER_BOUND)
    write_poly_with_type(group, constraint_binary.polynomial)

@add.static_method(ConstraintBinary)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    init_kwargs = hdf5_attributes.read_attributes(group, LOWER_BOUND, UPPER_BOUND)
    init_kwargs.update(polynomial=read_poly_with_type(group))
    return init_kwargs



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" package quark """

from .variable_mapping import VariableMapping
from .polynomial import Polynomial
from .poly_binary import PolyBinary
from .poly_ising import PolyIsing
from .constraint_binary import ConstraintBinary
from .objective import Objective
from .objective_terms import ObjectiveTerms
from .constrained_objective import ConstrainedObjective
from .solution import Solution
from .scip_model import ScipModel
from .hardware_adjacency import HardwareAdjacency
from .embedding import Embedding



# Copyright 2023 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the ScipModel with loaded ObjectiveTerms """

import os
import pytest

from quark import ScipModel, ConstrainedObjective
from quark.io import hdf5


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "constrained_objective.h5"
FILENAME_FIX = TEST_DIR + "testdata/"+ FILE


def test_init_with_load_hdf5():
    """ test ScipModel with hdf5 io (probably causing issues due to numpy arrays) """
    # TODO: set up other example without warnings
    with pytest.warns(UserWarning) as record:
        loaded_const_obj = hdf5.load(ConstrainedObjective, FILENAME_FIX, full_group_name="PFConstrainedObjective")
    assert len(record) == 2
    assert str(record[0].message).startswith("Experimental parsing")
    assert str(record[1].message).startswith("The two variables '('prime1', 1)' and '('prime2', 1)'")

    model = ScipModel.get_from_constrained_objective(loaded_const_obj)
    assert len(model.getConss()) == 18



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for Objective """

from abc import abstractmethod
from functools import partial
from warnings import warn

from .variable_mapping import VariableMapping
from .polynomial import Polynomial
from .poly_binary import PolyBinary
from .poly_ising import PolyIsing
from .utils import variables


POLYNOMIAL = 'polynomial'
NAME = 'name'

ERROR_POLYNOMIAL = "Objective's polynomial is neither Polynomial, PolyIsing nor PolyBinary"
ERROR_ISING      = "Only supported for Ising models"
ERROR_ZERO       = "Only supported for Ising models without linear terms (zero local fields)"
ERROR_VARIABLE   = "Cannot find variable '{}' in polynomial"
ERROR_IMPL = "Provide '{0}' on initialization or implement '_get_{0}' in inheriting 'ConstrainedObjective' class"
WARNING_NOTHING = "There is nothing in this Objective"
WARNING_DEPRECATION = "This way of instantiation is deprecated and will be removed soon, " \
                      "use method `get_from_instance` instead"


class Objective:
    """
    An objective is a function that shall be optimized in an optimization problem.
    Here we deal with polynomial objective functions for QUBO or Ising problems.
    Therefore, at this stage, there are no further constraints (anymore).
    """

    def __init__(self, polynomial=None, name=None, instance=None):
        """
        initialize Objective object

        :param (Polynomial) polynomial: the actual objective function
        :param (str or None) name: identifying name to differ between several objectives
        :param instance: instance object storing all data to construct the objective from
        """
        if instance:
            warn(WARNING_DEPRECATION, DeprecationWarning)
            self.polynomial = self._get_polynomial(instance)
            self.name = name or self._get_name(instance)
        else:
            self.polynomial = polynomial
            self.name = name

        super().__init__()
        self.check_consistency()

    def check_consistency(self):
        """ check if the data is consistent """
        if not isinstance(self.polynomial, Polynomial):
            raise ValueError(ERROR_POLYNOMIAL)
        if not self.polynomial:
            warn(WARNING_NOTHING)

    def __eq__(self, other):
        if isinstance(other, Objective):
            return self.name == other.name and self.polynomial == other.polynomial
        return False

    def __repr__(self):
        """ get nice string representation """
        return self.__class__.__name__ + f"({repr(self.polynomial)}, {repr(self.name)})"

    def __str__(self):
        """ get nice human-readable string representation of the objective """
        objective = f"min  {self.polynomial}\ns.t. "
        domains = sorted(set(variables.to_domain_string(var, self.polynomial.__class__.__name__)
                             for var in self.polynomial.variables))
        domains = "\n     ".join(domains)
        return objective + domains

    @classmethod
    def get_from_instance(cls, instance, name=None):
        """
        get the implemented inheriting Objective object based on the instance data

        :param (Instance or object) instance: An object of the implemented (inheriting) Instance class
        :param (str or None) name: the name of the resulting object
        :return: the initialized object
        """
        polynomial = cls._get_polynomial(instance)
        name = name or cls._get_name(instance)
        return cls(polynomial, name)

    def is_binary(self):
        """
        check if the objective's polynomial is binary

        :return: True if the polynomial is binary
        """
        return isinstance(self.polynomial, PolyBinary)

    def is_ising(self, inverted=None):
        """ check if the objective's polynomial is Ising

        :param (bool or None) inverted: if None, only checks for Ising without the inverted flag
                                        if True, additionally checks if the Ising polynomial is inverted,
                                        if False, additionally checks if the Ising polynomial is not inverted
        :return: True if the polynomial is Ising
        """
        if inverted is None:
            return isinstance(self.polynomial, PolyIsing)
        if inverted:
            return isinstance(self.polynomial, PolyIsing) and self.polynomial.is_inverted()
        return isinstance(self.polynomial, PolyIsing) and not self.polynomial.is_inverted()

    def apply_to_polynomial(self, func, new_name=None):
        """
        apply the function to the polynomial and thus extract another objective

        :param (callable) func: a mapping from the polynomial to a new one
        :param (str or None) new_name: name of the new objective, by default the original name is used
        :return: the new objective
        """
        new_name = new_name or self.name
        new_poly = func(self.polynomial)
        return Objective(new_poly, new_name)

    def to_binary(self, new_name=None):
        """
        convert this objective to a binary one

        :param (str or None) new_name: name of the new objective, by default the original name is used
        :return: the new Objective with binary polynomial
        """
        if self.is_binary() and not new_name:
            return self
        return self.apply_to_polynomial(PolyBinary.from_unknown_poly, new_name)

    def to_ising(self, invert=False, new_name=None):
        """
        convert this objective to an Ising one, with invert set to True the inverted Ising objective can be obtained

        :param (bool) invert: if True, in case the objective is binary it will be converted to inverted Ising,
                              in case the objective is Ising already it will be inverted
        :param (str or None) new_name: name of the new objective, by default the original name is used
        :return: the new Objective with Ising polynomial
        """
        if self.is_ising() and not invert and not new_name:
            return self
        return self.apply_to_polynomial(partial(PolyIsing.from_unknown_poly, invert=invert), new_name)

    def compact(self, new_name=None):
        """
        get a new Objective with a compact Polynomial

        :param (str or None) new_name: name of the new objective, by default the original name is used
        :return: the Objective with a compacted polynomial,
                 the VariableMapping of the original polynomial to new one (or None if it was already compact)
        """
        if self.polynomial.is_compact():
            return self, None
        return self.apply_to_polynomial(Polynomial.compact, new_name), VariableMapping(self.polynomial.variables)

    def break_symmetry_by_fixing_variable(self, variable=-1, value=1, new_name=None):
        """
        in case there are no linear terms (no local fields) in the Ising polynomial,
        the inverted variable assignment yields the same objective value as the non-inverted,
        we can break the symmetry explicitly by fixing one variable to a given value,
        this reduces the complexity by one without loss of generality

        :param (int or str or tuple) variable: variable to be fixed,
                                               by default -1 corresponding to the last in the sorted list
        :param (int) value: value to which the variable is fixed, by default 1
        :param (str or None) new_name: name of the new objective, by default the original name is used
        :return: the new Objective with one variable evaluated to the value
        """
        if not self.is_ising():
            raise ValueError(ERROR_ISING)
        if any(self.polynomial.coefficients_lists[1]):
            raise ValueError(ERROR_ZERO)
        if variable != -1 and variable not in self.polynomial.variables:
            raise ValueError(ERROR_VARIABLE.format(variable))
        if variable == -1:
            variable = self.polynomial.variables[-1]
        return self.apply_to_polynomial(partial(Polynomial.evaluate, var_assignments={variable: value}), new_name)

    @staticmethod
    @abstractmethod
    def _get_polynomial(instance):
        """
        get the objective polynomial from the instance data,
        needs to be implemented in subclasses for automatic generation

        :param instance: instance object storing all data to construct the objective polynomial from
        :return: the polynomial representing the objective function
        """
        raise NotImplementedError(ERROR_IMPL.format(POLYNOMIAL))

    @staticmethod
    def _get_name(instance):  # pylint: disable=unused-argument
        """
        get the name from the instance data,
        can be implemented in subclasses for automatic generation, is not enforced

        :param instance: instance object storing all data to construct the name from
        :return: the name of the object
        """
        return None



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the Solution """

import os
import pytest

from quark import Solution
from quark.io import hdf5


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "solution.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE

NAME = "test_solution"
NAME_EMPTY = "empty_solution"
NAME_NON_COMPACT = "non-compact_solution"


@pytest.fixture
def solutions():
    """ set up method """
    solving_time = 12
    solving_success = True
    solving_status = "optimal"
    objective_value = -12.1
    var_assignment = {0: 0, 1: 1, 2: 1, 3: 0, 4: 0, 5: 1, 6: 0, 7: 0, 8: 1}
    dual_gap = 1.1
    dual_bound = 0.1
    
    solution = Solution(var_assignments=var_assignment,
                        objective_value=objective_value,
                        solving_success=solving_success,
                        solving_status=solving_status,
                        solving_time=solving_time,
                        dual_gap=dual_gap,
                        dual_bound=dual_bound,
                        name=NAME)

    solution_empty = Solution(var_assignments={},
                              objective_value=objective_value,
                              solving_success=solving_success,
                              solving_status=solving_status,
                              solving_time=solving_time,
                              dual_gap=dual_gap,
                              dual_bound=dual_bound,
                              name=NAME_EMPTY)

    solution_nc = Solution(var_assignments={('x', 0, 1): 1, ('y', 3): 0},
                           objective_value=objective_value,
                           solving_success=solving_success,
                           solving_status=solving_status,
                           solving_time=solving_time,
                           dual_gap=dual_gap,
                           dual_bound=dual_bound,
                           name=NAME_NON_COMPACT)

    yield solution, solution_empty, solution_nc
        
def test_load_hdf5(solutions):
    """ test hdf5 io """
    solution, solution_empty, solution_nc = solutions

    loaded = hdf5.load(Solution, FILENAME_FIX, name=NAME)
    assert loaded == solution

    loaded = hdf5.load(Solution, FILENAME_FIX, name=NAME_EMPTY)
    assert loaded == solution_empty

    loaded = hdf5.load(Solution, FILENAME_FIX, name=NAME_NON_COMPACT)
    assert loaded == solution_nc

def test_io_hdf5(solutions):
    """ test hdf5 io round trip """
    solution, solution_empty, solution_nc = solutions

    hdf5.save(solution, FILENAME_TMP)
    loaded = hdf5.load(Solution, FILENAME_TMP, name=NAME)
    assert loaded == solution

    hdf5.save(solution_empty, FILENAME_TMP)
    loaded = hdf5.load(Solution, FILENAME_TMP, name=NAME_EMPTY)
    assert loaded == solution_empty

    hdf5.save(solution_nc, FILENAME_TMP)
    loaded = hdf5.load(Solution, FILENAME_TMP, name=NAME_NON_COMPACT)
    assert loaded == solution_nc

    os.remove(FILENAME_TMP)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the Polynomial """

import ast

from quark import Polynomial
from . import add, hdf5_datasets, hdf5_attributes, txt
from ._variable_mapping import write_variables, read_variables


FLAT_POLYNOMIAL = "flat_polynomial"


add.save_load_exists(Polynomial)

@add.self_method(Polynomial)
def write_hdf5(polynomial, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (Polynomial) polynomial: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    polynomial.sort_entries()
    if polynomial.is_flat():
        flat_polynomial = polynomial
    else:
        flat_polynomial = polynomial.compact()
        write_variables(polynomial.variables, group)
    hdf5_datasets.write_dataset(group, FLAT_POLYNOMIAL, dataset=flat_polynomial)
    hdf5_attributes.write_attributes(group, **polynomial.coefficients_info)

@add.static_method(Polynomial)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    flat_poly_dict_np = hdf5_datasets.read_dataset(group, FLAT_POLYNOMIAL)
    flat_poly_dict = {tuple(int(i) for i in k): v for k, v in flat_poly_dict_np.items()}
    poly = Polynomial(flat_poly_dict)

    composite_variables = read_variables(group)
    if composite_variables:
        poly = poly.replace_variables(composite_variables)
    return {"d": poly, "variable_tuples_already_formatted": True}


# Adding IO methods to Polynomial that call the corresponding ones of txt
add.save_load_exists(Polynomial, txt, "txt")

@add.self_method(Polynomial)
def write_txt(polynomial, txt_file):
    """
    write data in opened text file

    :param (Polynomial) polynomial: the polynomial to be saved
    :param (file) txt_file: the file to save the data in
    """
    txt_file.write(super(Polynomial, polynomial).__repr__().replace(", (", ",\n ("))

@add.static_method(Polynomial)
def read_txt(lines):
    """
    read data from opened text file

    :param (list[str]) lines: the loaded lines of the text file to load the data from
    :return: the data as keyword arguments
    """
    line = "".join(lines).replace("\n", "")
    return {"d": ast.literal_eval(line)}



""" setup for project quark """

from setuptools import setup, find_packages


setup(name = 'quark',
      version = '1.2',
      url = 'https://gitlab.com/quantum-computing-software/quark/',
      author = 'DLR-SC',
      author_email = 'qc-software@dlr.de',
      python_requires = '>=3.11',
      description = 'This is a software package to support the mapping of combinatorial optimization problems to quantum computing interfaces via QUBO and Ising problems.',
      packages = find_packages(include=['quark', 'quark.io', 'quark.testing', 'quark.utils']),
      zip_safe = False)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Tests for the solution class """

import pytest

from quark import Solution
from quark.solution import binary_to_ising, ising_to_binary, invert_ising


X = "x"

NAME = "test_solution"

INVALID_SOL = Solution(var_assignments={1: -1, 2: 0, 3: 1}, name=NAME)
UNDEFINED_SOL = Solution(var_assignments={1: 1, 2: 1, 3: 1}, name=NAME)
COMPACT_SOL = Solution(var_assignments={0: 1, 1: 0, 2: 1}, solving_status="timeout", name=NAME)

VAR_ASSIGNMENTS_ISING_SOL = {(X, 0, 0): 1, (X, 4, 5): -1, (X, 3, 2): 1}
VAR_ASSIGNMENTS_ISING_INV_SOL = {(X, 0, 0): -1, (X, 4, 5): 1, (X, 3, 2): -1}
VAR_ASSIGNMENTS_BINARY_SOL = {(X, 4, 5): 0, (X, 3, 2): 1, (X, 0, 0): 1}

ISING_SOL = Solution(var_assignments=VAR_ASSIGNMENTS_ISING_SOL, solving_status="optimal", name=NAME)
INV_ISING_SOL = Solution(var_assignments=VAR_ASSIGNMENTS_ISING_INV_SOL, solving_status="optimal", name=NAME)
BINARY_SOL = Solution(var_assignments=VAR_ASSIGNMENTS_BINARY_SOL, solving_status="timeout", name=NAME)


def test_init():
    """ test initialization """
    solution = Solution([-1, 1, -1, 1], solving_time=60)
    assert dict(solution) == {0: -1, 1: 1, 2: -1, 3: 1}

def test_has_only_binary_values():
    """ test function for checking if solution has only binary values """
    assert BINARY_SOL.has_only_binary_values()
    assert UNDEFINED_SOL.has_only_binary_values()
    assert not ISING_SOL.has_only_binary_values()

def test_has_only_ising_values():
    """ test function for checking if solution has only ising values """
    assert ISING_SOL.has_only_ising_values()
    assert UNDEFINED_SOL.has_only_ising_values()
    assert not BINARY_SOL.has_only_ising_values()

def test_to_ising():
    """ test conversion of solution to ising form """
    with pytest.raises(ValueError, match="Solution is neither binary nor ising"):
        assert INVALID_SOL.to_ising()
    # solutions should be identical
    assert ISING_SOL.to_ising() == ISING_SOL

    # just invert solution
    inv_ising_sol = ISING_SOL.to_ising(invert=True)
    assert inv_ising_sol == INV_ISING_SOL

    binary_2_ising = BINARY_SOL.to_ising()
    # just checking for the var_assignments
    assert binary_2_ising == VAR_ASSIGNMENTS_ISING_SOL
    # also checking for additional attribute
    assert binary_2_ising.solving_status == BINARY_SOL.solving_status
    assert binary_2_ising.name == ISING_SOL.name

    binary_2_ising = BINARY_SOL.to_ising(invert=True)
    # just checking for the var_assignments
    assert binary_2_ising == VAR_ASSIGNMENTS_ISING_INV_SOL
    # also checking for additional attribute
    assert binary_2_ising.solving_status == BINARY_SOL.solving_status
    assert binary_2_ising.name == INV_ISING_SOL.name
    assert binary_2_ising != INV_ISING_SOL

def test_to_binary():
    """ test conversion of solution to binary form """
    with pytest.raises(ValueError, match="Solution is neither binary nor ising"):
        assert INVALID_SOL.to_binary()
    # var_assignments should be identical
    assert BINARY_SOL.to_binary() == BINARY_SOL
    # also checking for additional attribute
    assert BINARY_SOL.to_binary().solving_status == BINARY_SOL.solving_status

    ising_2_binary = ISING_SOL.to_binary()
    # just checking for the var_assignments
    assert ising_2_binary == VAR_ASSIGNMENTS_BINARY_SOL
    # also checking for additional attribute
    assert ising_2_binary.solving_status == ISING_SOL.solving_status

    ising_2_binary = ISING_SOL.to_binary(is_inverted=True)
    # just checking for the var_assignments
    assert ising_2_binary == {(X, 0, 0): 0, (X, 4, 5): 1, (X, 3, 2): 0}
    # also checking for additional attribute
    assert ising_2_binary.solving_status == ISING_SOL.solving_status
    assert ising_2_binary != [(X, 0, 0), (X, 4, 5), (X, 3, 2)]

def test_decompact():
    """ test reversion of compact """
    variables = [(X, 0, 0), (X, 4, 5), (X, 3, 2)]
    decompact_sol = COMPACT_SOL.decompact(variables)
    assert decompact_sol == BINARY_SOL
    assert decompact_sol.solving_status == BINARY_SOL.solving_status

    with pytest.raises(ValueError, match="Solution is not compact"):
        BINARY_SOL.decompact(variables)
    with pytest.raises(ValueError, match="Solution is not compact"):
        UNDEFINED_SOL.decompact(variables)

    too_less = [(X, 0, 0), (X, 4, 5)]
    with pytest.raises(ValueError, match="Number of provided variables is different than of contained variables"):
        COMPACT_SOL.decompact(too_less)

    wrong = set(variables)
    with pytest.raises(ValueError, match="Dictionary of variables or VariableMapping is expected"):
        # noinspection PyTypeChecker
        COMPACT_SOL.decompact(wrong)

    with pytest.raises(ValueError, match="Solution is empty, there are no variables to be replaced"):
        Solution({}).decompact(variables)

def test_replace_variables():
    """ test variable replacement in solution """
    replacement = {(X, 0, 0): "a", (X, 4, 5): "b", (X, 3, 2): "c"}
    exp_var_assigment = {"a": 1, "b": 0, "c": 1}
    replaced_sol = BINARY_SOL.replace_variables(replacement)
    assert replaced_sol == exp_var_assigment
    assert replaced_sol.solving_status == "timeout"

    too_less = {(X, 0, 0): "a", (X, 4, 5): "b"}
    exp_var_assigment = {"a": 1, "b": 0, (X, 3, 2): 1}
    replaced_sol = BINARY_SOL.replace_variables(too_less, check_all=False)
    assert replaced_sol == exp_var_assigment

    with pytest.raises(ValueError, match="Variable '.*' does not have a mapping to a new variable"):
        BINARY_SOL.replace_variables(too_less)

    replacement = [(X, 0, 0), (X, 4, 5), (X, 3, 2)]
    replaced_sol = COMPACT_SOL.replace_variables(replacement)
    assert replaced_sol == BINARY_SOL

def test_binary_ising_conversion():
    """ test conversion of binary and ising variable assignments """
    assert binary_to_ising(VAR_ASSIGNMENTS_BINARY_SOL) == VAR_ASSIGNMENTS_ISING_SOL
    assert binary_to_ising(VAR_ASSIGNMENTS_BINARY_SOL, invert=True) == VAR_ASSIGNMENTS_ISING_INV_SOL
    assert ising_to_binary(VAR_ASSIGNMENTS_ISING_SOL) == VAR_ASSIGNMENTS_BINARY_SOL
    assert ising_to_binary(VAR_ASSIGNMENTS_ISING_INV_SOL, is_inverted=True) == VAR_ASSIGNMENTS_BINARY_SOL
    assert invert_ising(VAR_ASSIGNMENTS_ISING_INV_SOL) == VAR_ASSIGNMENTS_ISING_SOL
    assert invert_ising(VAR_ASSIGNMENTS_ISING_SOL) == VAR_ASSIGNMENTS_ISING_INV_SOL

def test_repr():
    """ test string generation """
    assert repr(BINARY_SOL) == "Solution({('x', 0, 0): 1, ('x', 3, 2): 1, ('x', 4, 5): 0}, None, 'test_solution')"

def test_str():
    """ test string generation """
    assert str(BINARY_SOL) == "x_0_0 \t= 1\nx_3_2 \t= 1\nx_4_5 \t= 0"

def test_sort_entries():
    """ test sorting """
    solution = Solution(VAR_ASSIGNMENTS_BINARY_SOL)
    assert list(solution.keys()) == [(X, 4, 5), (X, 3, 2), (X, 0, 0)]
    solution.sort_entries()
    assert list(solution.keys()) == [(X, 0, 0), (X, 3, 2), (X, 4, 5)]



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ExampleInstance """

from quark.io import Instance, hdf5_datasets
from . import example_constrained_objective


EDGES = "edges"
DENSITY = "density"
COLORS = "colors"

ERROR_PREFIX = "Instance is not consistent: "
ERROR_EDGES  = "No edges given"
ERROR_COLORS = "No colors given"
ERROR_TWO    = "Edges need to have two nodes"


class ExampleInstance(Instance):
    """ An example implementation of an Instance for testing based on the MaxColorableSubgraph problem """

    def __init__(self, edges, colors):
        """
        an instance contains

        :param (list or set) edges: undirected edges of the graph -> define the set of nodes
        :param (list or set or int) colors: set of colors or int defining range of colors
        """
        self.edges = set(tuple(sorted(edge)) for edge in edges)
        self.nodes = set(node for edge in self.edges for node in edge)
        self.colors = set(range(colors)) if isinstance(colors, int) else set(colors)
        super().__init__()

    def check_consistency(self):
        """ check if the data is consistent """
        if len(self.edges) == 0:
            raise ValueError(ERROR_PREFIX + ERROR_EDGES)
        if len(self.colors) == 0:
            raise ValueError(ERROR_PREFIX + ERROR_COLORS)
        if not all(len(edge) == 2 for edge in self.edges):
            raise ValueError(ERROR_PREFIX + ERROR_TWO)

    def get_name(self):
        """ get an expressive name representing the instance """
        density = round(2 * len(self.edges) / (len(self.nodes) * (len(self.nodes) - 1)), 2)
        return "_".join([self.__class__.__name__, DENSITY, f"{density}", COLORS, f"{len(self.colors)}"])

    def get_constrained_objective(self, name=None):
        """
        get the constrained objective object based on this instance data

        :param (str or None) name: the name of the object
        :return: the constrained objective
        """
        return example_constrained_objective.ExampleConstrainedObjective.get_from_instance(self, name)

    def write_hdf5(self, group):
        """
        write data in attributes and datasets in the opened group of the HDF5 file

        :param (h5py.Group) group: the group to store the data in
        """
        hdf5_datasets.write_datasets(group, self, EDGES, COLORS)

    @staticmethod
    def read_hdf5(group):
        """
        read data from attributes and datasets of the opened group of the HDF5 file

        :param (h5py.Group) group: the group to read the data from
        :return: the data as keyword arguments
        """
        kwargs = hdf5_datasets.read_datasets(group, EDGES, COLORS)
        kwargs[EDGES] = [tuple(edge) for edge in kwargs[EDGES]]
        kwargs[COLORS] = list(kwargs[COLORS])
        return {**kwargs}



# Copyright 2025 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for namedtuple variables """

from collections import namedtuple

from .poly_binary import PolyBinary
from .poly_ising import PolyIsing


def var_binary_namedtuple(name, *field_names):
    """
    get a new subclass of tuple representing binary variables with the same tuple structure

    with `x = var_binary_namedtuple("x", "node", "index")` you can get the constructor and then
    use it for a specific variable with `x("a", 1)`, which gives you the namedtuple `Bin(var='x', node='a', index=1)`

    note that the addition, multiplication, etc. of these namedtuple variables is rather to get familiar with our
    polynomials and variable tuples, it is not optimized for performance,
    thus we do not recommend the usage on large scale
    """

    class Bin(namedtuple("Bin", ("var",) + field_names)):
        """ workaround to define the namedtuple in a certain way """

        def __new__(cls, *args, **kwargs):
            # the name is inserted for this variable by default
            return super(Bin, cls).__new__(cls, name, *args, **kwargs)

        def __add__(self, other):

            if isinstance(other, Bin):
                other = PolyBinary({(other,) : 1})
            return PolyBinary({(self,) : 1}) + other

        def __radd__(self, other):
            return PolyBinary({(self,) : 1}) + other

        def __sub__(self, other):
            if isinstance(other, Bin):
                other = PolyBinary({(other,) : 1})
            return PolyBinary({(self,) : 1}) - other

        def __rsub__(self, other):
            return other - PolyBinary({(self,) : 1})

        def __mul__(self, other):
            if isinstance(other, Bin):
                other = PolyBinary({(other,) : 1})
            return other * PolyBinary({(self,) : 1})

        def __rmul__(self, other):
            return other * PolyBinary({(self,) : 1})

        def __truediv__(self, other):
            return PolyBinary({(self,) : 1}) / other

        def __pow__(self, exponent):
            return PolyBinary({(self,) : 1}) ** exponent

    return Bin


def var_ising_namedtuple(name, *field_names):
    """
    get a new subclass of tuple representing Ising variables with the same tuple structure

    with `s = var_binary_namedtuple("s", "node", "index")` you can get the constructor and then
    use it for a specific variable with `s("a", 1)`, which gives you the namedtuple `Ising(var='s', node='a', index=1)`

    note that the addition, multiplication, etc. of these namedtuple variables is rather to get familiar with our
    polynomials and variable tuples, it is not optimized for performance,
    thus we do not recommend the usage on large scale
    """

    class Ising(namedtuple("Ising", ("var",) + field_names)):
        """ workaround to define the namedtuple in a certain way """

        def __new__(cls, *args, **kwargs):
            # the name is inserted for this variable by default
            return super(Ising, cls).__new__(cls, name, *args, **kwargs)

        def __add__(self, other):
            if isinstance(other, Ising):
                other = PolyIsing({(other,) : 1})
            return PolyIsing({(self,) : 1}) + other

        def __radd__(self, other):
            return PolyIsing({(self,) : 1}) + other

        def __sub__(self, other):
            if isinstance(other, Ising):
                other = PolyIsing({(other,) : 1})
            return PolyIsing({(self,) : 1}) - other

        def __rsub__(self, other):
            return other - PolyIsing({(self,) : 1})

        def __mul__(self, other):
            if isinstance(other, Ising):
                other = PolyIsing({(other,) : 1})
            return other * PolyIsing({(self,) : 1})

        def __rmul__(self, other):
            return other * PolyIsing({(self,) : 1})

        def __truediv__(self, other):
            return PolyIsing({(self,) : 1}) / other

        def __pow__(self, exponent):
            return PolyIsing({(self,) : 1}) ** exponent

    return Ising



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ConstraintObjective """

from abc import abstractmethod
from warnings import warn

from .variable_mapping import VariableMapping
from .polynomial import _try_sorted
from .poly_binary import PolyBinary, get_all_reduction_penalty_terms
from .constraint_binary import get_reduction_constraints
from .constraint_binary import ConstraintBinary
from .objective_terms import ObjectiveTerms
from .utils import variables


OBJECTIVE      = "objective"
REDUCTION      = "reduction"
OBJECTIVE_POLY = "objective_poly"
CONSTRAINTS    = "constraints"

ERROR_INIT = "Currently only binary objective functions are supported for the ConstrainedObjective"
ERROR_IMPL = "Provide '{0}' on initialization or implement '_get_{0}' in inheriting 'ConstrainedObjective' class"
ERROR_UNFULFILLABLE = "Unfulfillable set of constraints"
ERROR_TYPE = "Objective polynomial's and constraints' variable types do not match"
WARNING_NOTHING = "There is nothing in this ConstrainedObjective"
WARNING_DUPLICATED = "The ConstrainedObjective contains {} redundant constraint(s)"
WARNING_DEPRECATION = "This way of instantiation is deprecated and will be removed soon, " \
                      "use method `get_from_instance` instead"


class ConstrainedObjective(dict):
    """
    An objective function combined with usually multiple constraints, restricting the set of feasible solutions,
    together provide an optimization problem.

    This class either serves as a container for a pre-calculated objective polynomial with constraints
    or as a base class for the implementation of methods providing the objective polynomial and the constraints.

    (It is analogue to ObjectiveTerms but with storing constraints.)
    """

    def __init__(self, objective_poly=None, constraints=None, name=None, instance=None):
        """
        initialize a ConstrainedObjective object,
        either provide objective_poly and constraints to store them or
        provide instance for implemented subclass to construct the objective_poly and constraints from the instance

        :param (PolyBinary or None) objective_poly: polynomial representing the objective function
        :param (dict or None) constraints: dictionary with names to constraints
        :param (str or None) name: identifying name to differ between several constrained objectives
        :param instance: instance object storing all data to construct the constrained objective from
        """
        if instance:
            warn(WARNING_DEPRECATION, DeprecationWarning)
            self.objective_poly = self._get_objective_poly(instance)
            self.name = name or self._get_name(instance)
            super().__init__(self._get_constraints(instance))
        else:
            self.objective_poly = objective_poly or PolyBinary()
            self.name = name
            super().__init__(constraints or {})

        self.check_consistency()

    def check_consistency(self):
        """ check if the data is consistent """
        if not isinstance(self.objective_poly, PolyBinary):
            raise ValueError(ERROR_INIT)
        if not self.objective_poly and not self:
            warn(WARNING_NOTHING)

        redundant_cons = self.get_redundant_constraints()
        if redundant_cons:
            warn(WARNING_DUPLICATED.format(len(redundant_cons)))
            # TODO: detect more contradicting constraints? -> preprocess

        # check the variables from the objective polynomial and the constraints
        _ = self.get_all_variables()

    def __eq__(self, other):
        if isinstance(other, ConstrainedObjective):
            if super().__eq__(other):
                return self.objective_poly == other.objective_poly and self.name == other.name
        return False

    def __ne__(self, other):
        # needs to be implemented otherwise would just use dict comparison
        return not self.__eq__(other)

    def __repr__(self):
        """ get nice string representation """
        return self.__class__.__name__ + f"({repr(self.objective_poly)}, {super().__repr__()}, {repr(self.name)})"

    def __str__(self):
        """ get nice human-readable string representation of the constrained objective """
        objective = f"min  {self.objective_poly}\ns.t. "
        domains = sorted(set(variables.to_domain_string(var) for var in self.get_all_variables()))

        lens_lower_bounds = [len(str(constraint.lower_bound)) for constraint in self.values()]
        max_len_lower_bound = max(lens_lower_bounds, default=0)
        const_prefixes = [" " * (max_len_lower_bound - len_lower_bound) for len_lower_bound in lens_lower_bounds]
        constraints = [prefix + constraint.__str__(revert_eq=True)
                       for prefix, constraint in zip(const_prefixes, self.values())
                       if constraint.is_equality_constraint()]
        constraints += [prefix + str(constraint)
                        for prefix, constraint in zip(const_prefixes, self.values())
                        if not constraint.is_equality_constraint()]
        constraints = [constraint for prefix, constraint in zip(const_prefixes, constraints)]

        subject_to = ",\n     ".join(constraints + domains)
        return objective + subject_to

    @classmethod
    def get_from_instance(cls, instance, name=None):
        """
        get the implemented inheriting ConstrainedObjective object based on the instance data

        :param (Instance or object) instance: An object of the implemented (inheriting) Instance class
        :param (str or None) name: the name of the resulting object
        :return: the initialized object
        """
        objective_poly = cls._get_objective_poly(instance)
        constraints = cls._get_constraints(instance)
        name = name or cls._get_name(instance)
        return cls(objective_poly, constraints, name)

    def get_broken_constraints(self, var_assignment):
        """
        get the names of the constraints that the variable assignment does not fulfill

        :param (dict or list) var_assignment: mapping from variables to values
        :return: the names of the constraints that are not fulfilled
        """
        return [name for name, constraint in self.items() if not constraint.check_validity(var_assignment)]

    def get_redundant_constraints(self):
        """
        get the redundant constraints

        :return: the names of redundant constraints
        """
        non_redundant_cons = {}
        for name, constraint in self.items():
            if constraint not in non_redundant_cons.values():
                non_redundant_cons[name] = constraint
        redundant_cons = set(self.keys()) - set(non_redundant_cons.keys())
        return redundant_cons

    def remove_redundant_constraints(self):
        """
        remove redundant constraints

        :return: a ConstrainedObjective without redundant constraints
        """
        redundant_constraints = self.get_redundant_constraints()
        if redundant_constraints:
            constraints = {name: constraint.copy() for name, constraint in self.items()
                           if not name in redundant_constraints}
            return self.__class__(self.objective_poly.copy(), constraints)
        return self

    def check_validity(self, var_assignment):
        """
        check if the variable assignment fulfills the constraints

        :param (dict or list) var_assignment: mapping from variables to values
        :return: True if all constraints are fulfilled
        """
        return not self.get_broken_constraints(var_assignment)

    def is_compact(self):
        """
        determine if the ConstraintObjective is compact
        :return: bool
        """
        return variables.are_consecutive(self.get_all_variables())

    def compact(self, new_name=None):
        """
        get a new ConstrainedObjective with a compact Polynomial for the
        objective and constraints

        :param (str or None) new_name: name of the new objective, by default
                                       the original name is used
        :return: the ConstrainedObjective with a compacted polynomial,
                 the VariableMapping of the original polynomial to new one
                 (or None if it was already compact)
        """
        if self.is_compact():
            return self, None

        new_name = new_name or self.name
        all_variables = self.get_all_variables()
        compact_objective_poly = self.objective_poly.replace_variables_by_ordering(all_variables)
        compact_constraints = {}
        for name, constraint in self.items():
            compact_poly = constraint.polynomial.replace_variables_by_ordering(all_variables)
            compact_constraints[name] = ConstraintBinary(compact_poly, constraint.lower_bound, constraint.upper_bound)
        return self.__class__(compact_objective_poly, compact_constraints, new_name), VariableMapping(all_variables)

    def get_reductions(self, max_degree=1, use=None, force=False, reduction_strategy=None):
        """
        get new constrained objective where all appearing polynomials are reduced if the degrees are too large

        :param (int or None) max_degree: the maximal degree of the final polynomial,
                                         if None only the given reductions are applied
        :param (list[tuples] or None) use: the reductions which shall be used,
                                           as a list with tuples in the format (var1, var2, reduction_var)
        :param (bool) force: if True also reduce quadratic polynomials to linear that could remain for the QUBO problem
        :param (str) reduction_strategy: the method how the reduction pairs are found, by default the fastest version
        :return: the reduced constrained objective object, the applied reductions
        """
        # TODO: combine all polynomials to get corresponding possibly better reductions for full object
        reduced_objective_poly, reductions = self.objective_poly.reduce(1 if force else max(2, max_degree), use=use,
                                                                        reduction_strategy=reduction_strategy)
        use = use or []
        reduced_constraints = {}
        for name, constraint in self.items():
            reduced_constraint, further_reductions = constraint.get_reductions(max_degree, use + reductions, force,
                                                                               reduction_strategy=reduction_strategy)
            reductions += further_reductions
            if reduced_constraint:
                reduced_constraints[name] = reduced_constraint

        # go through constraints again and apply reductions found in a later step also in earlier processed constraints
        reductions = sorted(set(reductions))
        for name, constraint in reduced_constraints.items():
            reduced_constraints_single = constraint.reduce(None, use=reductions, reduction_strategy=reduction_strategy,
                                                           reduced_constraint_name=name)
            assert len(reduced_constraints_single) == 1  # no reduction constraints should be added
            reduced_constraints.update(reduced_constraints_single)

        return reduced_objective_poly, reduced_constraints, reductions

    def reduce(self, max_degree=1, use=None, force=False, new_name=None, reduction_strategy=None):
        """
        get new constrained objective where all appearing polynomials are reduced if the degrees are too large
        and the corresponding reduction constraints are added

        :param (int or None) max_degree: the maximal degree of the final polynomial,
                                         if None only the given reductions are applied
        :param (list[tuples] or None) use: the reductions which shall be used,
                                           as a list with tuples in the format (var1, var2, reduction_var)
        :param (bool) force: if True also reduce quadratic polynomials to linear that could remain for the QUBO problem
        :param (str or None) new_name: name of the new constrained objective, by default the original name is used
        :param (str) reduction_strategy: the method how the reduction pairs are found, by default the fastest version
        :return: the reduced constrained objective object, the applied reductions
        """
        reduced_obj_poly, reduced_constraints, reductions = self.get_reductions(max_degree, use, force,
                                                                                reduction_strategy=reduction_strategy)
        for reduction in reductions:
            reduced_constraints.update(get_reduction_constraints(*reduction, max_degree=max_degree))

        new_name = new_name or self.name
        return self.__class__(reduced_obj_poly, reduced_constraints, new_name)

    def get_objective_terms(self, objective_name=OBJECTIVE, combine_prefixes=(REDUCTION,), name=None,
                            reduction_strategy=None, check_special_constraints=True):
        """
        get an ObjectiveTerms object with the objective as one term
        and automatically generated penalty terms for each constraint

        :param (str) objective_name: name of the term that corresponds to the objective
        :param (tuple or str) combine_prefixes: prefixes in the names of the constraints whose corresponding
                                                objective terms shall be combined to a single objective term,
                                                by default all terms starting with "reduction" are combined
        :param (str or None) name: name of the objective terms, by default the name of the constrained objective is used
        :param (str) reduction_strategy: the method how the reduction pairs are found, by default the fastest version
        :param (bool) check_special_constraints: whether special constraints shall be considered
        :return: the ObjectiveTerms object
        """
        # TODO: catch constraints setting two variables to be equal and use polynomial.evaluate -> preprocess?
        # TODO: option to get objective terms without automatic reduction?
        reduced_obj_poly, reduced_constraints, reductions = self.get_reductions(reduction_strategy=reduction_strategy)

        objective_terms = {}
        for constraint_name, constraint in reduced_constraints.items():
            objective_terms.update(constraint.get_penalty_terms(constraint_name,
                                                                check_special_constraints=check_special_constraints))
        # after the reduction there should be a single penalty term for each constraint
        assert len(reduced_constraints) == len(objective_terms)

        reduction_penalty_terms = get_all_reduction_penalty_terms(reductions)
        objective_terms.update(reduction_penalty_terms)
        constraint_terms_names = list(objective_terms.keys())

        if not reduced_obj_poly == 0:
            # only in case the objective polynomial is not empty we do want to add it as an additional term
            objective_terms.update({objective_name: reduced_obj_poly})

        name = name or self.name
        obj = ObjectiveTerms(objective_terms=objective_terms, constraint_terms_names=constraint_terms_names, name=name)
        if combine_prefixes:
            combine_prefixes = combine_prefixes if isinstance(combine_prefixes, tuple) else (combine_prefixes,)
            return obj.combine_prefixes(*combine_prefixes)
        return obj

    def get_all_variables(self):
        """
        get all variables from all polynomials

        :return: the sorted list of all appearing variables
        """
        all_vars = set(self.objective_poly.variables).union(*(const.polynomial.variables for const in self.values()))
        try:
            _ = variables.get_common_type(all_vars)
        except TypeError as te:
            assert "Expected variable type" in te.args[0]
            raise TypeError(ERROR_TYPE) from te
        return _try_sorted(all_vars)

    @staticmethod
    @abstractmethod
    def _get_objective_poly(instance):
        """
        get the objective polynomial from the instance data,
        needs to be implemented in subclasses for automatic generation

        :param instance: instance object storing all data to construct the objective polynomial from
        :return: the polynomial representing the objective function
        """
        raise NotImplementedError(ERROR_IMPL.format(OBJECTIVE_POLY))

    @staticmethod
    @abstractmethod
    def _get_constraints(instance):
        """
        get the constraints from the instance data,
        needs to be implemented in subclasses for automatic generation

        :param instance: instance object storing all data to construct the constraints from
        :return: the dictionary with names to constraints
        """
        raise NotImplementedError(ERROR_IMPL.format(CONSTRAINTS))

    @staticmethod
    def _get_name(instance):  # pylint: disable=unused-argument
        """
        get the name from the instance data,
        can be implemented in subclasses for automatic generation, is not enforced

        :param instance: instance object storing all data to construct the name from
        :return: the name of the object
        """
        return None



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO with txt files """

import os
import pytest

from quark import Polynomial
from quark.io import txt


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "polynomial.txt"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE

POLY = Polynomial({(): 5.5, (("x", 1),): 2.5, (("x", 2), ("y", 0, 0)): 2.0})


def test_save():
    """ testing if the polynomial is correctly saved to the file """
    txt.save(POLY, FILENAME_TMP)
    with open(FILENAME_TMP) as txt_file:
        loaded_lines = txt_file.readlines()
    expected_lines = ["Polynomial\n", "{(): 5.5,\n", " (('x', 1),): 2.5,\n", " (('x', 2),\n", " ('y', 0, 0)): 2.0}\n"]
    assert loaded_lines == expected_lines
    os.remove(FILENAME_TMP)

def test_load():
    """ testing if the polynomial is correctly loaded from the file """
    loaded_poly = txt.load(Polynomial, FILENAME_FIX)
    assert loaded_poly == POLY

def test_exists():
    """ testing if the polynomial is correctly found in the file """
    wrong_filename = TEST_DIR + "testdata/poly_ising.txt"
    assert txt.exists(Polynomial, FILENAME_FIX)
    assert not txt.exists(Polynomial, wrong_filename)

def test_save_load_several():
    """ testing the loading and saving of several polynomials in one file """
    poly1 = Polynomial({(1, 2, 3): 100})
    poly2 = Polynomial({(1, 2): 1, (2, 3): 2, (3, 4): 3})

    txt.save(POLY, FILENAME_TMP)
    txt.save(poly1, FILENAME_TMP, "a")
    txt.save(poly2, FILENAME_TMP, "a")

    with open(FILENAME_TMP) as txt_file:
        loaded_lines = txt_file.readlines()

    exp_lines = ["Polynomial\n", "{(): 5.5,\n", " (('x', 1),): 2.5,\n", " (('x', 2),\n", " ('y', 0, 0)): 2.0}\n", "\n",
                 "Polynomial\n", "{(1, 2, 3): 100}\n", "\n",
                 "Polynomial\n", "{(1, 2): 1,\n", " (2, 3): 2,\n", " (3, 4): 3}\n"]
    assert loaded_lines == exp_lines

    loaded_poly0 = txt.load(Polynomial, FILENAME_TMP)
    loaded_poly1 = txt.load(Polynomial, FILENAME_TMP, index=1)
    loaded_poly2 = txt.load(Polynomial, FILENAME_TMP, index=2)

    assert loaded_poly0 == POLY
    assert loaded_poly1 == poly1
    assert loaded_poly2 == poly2

    with pytest.raises(ValueError, match="The file does not contain so much objects of type 'Polynomial'"):
        txt.load(Polynomial, FILENAME_TMP, index=3)

    with pytest.raises(ValueError, match="The file does not contain any object of type 'Polynomial'"):
        wrong_filename = TEST_DIR + "testdata/poly_ising.txt"
        txt.load(Polynomial, wrong_filename)

    os.remove(FILENAME_TMP)



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the variable utility functions """

import pytest

from quark import PolyIsing
from quark.utils.variables import get_common_type, get_type, check_type_against, are_consecutive, to_string, \
                                  replace_strs_by_ints, to_domain_string


def test_get_common_type():
    """ test finding the type of all variables """
    assert get_common_type([]) is None
    assert get_common_type(["x", "y"]) is str

    with pytest.raises(TypeError, match="Expected variable type 'tuple', but got 'str'"):
        get_common_type([(1, 4), "x"])

def test_get_type():
    """ test finding the type """
    assert get_type(("x", 1)) is tuple

    with pytest.raises(TypeError, match="Variable 'True' has invalid type 'bool'"):
        get_type(True)

    with pytest.raises(ValueError, match="should only contain ints and strings"):
        get_type(("x", 1.4))

def test_check_type_against():
    """ test check whether type is the expected one """
    with pytest.raises(TypeError, match="Expected variable type 'tuple', but got 'str'"):
        check_type_against("x", tuple)

def test_are_consecutive():
    """ test checking of consecutive variables """
    assert are_consecutive([0, 1, 2, 3])
    assert are_consecutive([1, 0, 1, 2, 3])
    assert not are_consecutive([1, 2, 3])
    assert not are_consecutive([1, 0, 4, 0])
    assert not are_consecutive([(0,), (1,), (2,), (3,)])
    assert not are_consecutive([(0,), "x"])

def test_to_string():
    """ test string creation """
    assert to_string(("s", 1, 2)) == "s_1_2"
    assert to_string((1, 2)) == "x_1_2"
    assert to_string(1) == "x1"
    assert to_string(1, PolyIsing) == "s1"

def test_to_domain_string():
    """ test string creation """
    assert to_domain_string(("s", 1, 2)) == "s_*_* in {0, 1}"
    assert to_domain_string((1, 2)) == "x_*_* in {0, 1}"
    assert to_domain_string(1) == "x* in {0, 1}"
    assert to_domain_string((1, 2), "PolyIsing") == "s_*_* in {-1, 1}"
    assert to_domain_string(1, PolyIsing) == "s* in {-1, 1}"

def test_replace_strs_by_ints():
    """ test replacement of string parts """
    expected_tuples = [(4, 5, 1), (6, 2, 3)]
    expected_int_to_str = {4 : "x", 5 : "y", 6 : "z"}
    tuples, int_to_str = replace_strs_by_ints([("x", "y", 1), ("z", 2, 3)])
    assert tuples == expected_tuples
    assert int_to_str == expected_int_to_str
